# runserver.py - command to serve Updian via Flask's builtin web server
#
#  Copyright (c) 2007-2013 RobHost GmbH <support@robhost.de>
#
#  Author: Dirk Dankhoff <dd@robhost.de>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA
from __future__ import print_function

import socket
import sys

from updian.frontend import app

params = dict(debug=False)

try:
    ip = sys.argv[1]
    if ip.count('.') != 3:
        raise socket.error()
    socket.inet_aton(ip)
    params['host'] = ip
except socket.error:
    print('ERROR: Given IPv4 address is invalid.', file=sys.stderr)
except IndexError:
    pass

app.run(**params)



# fabric_utils.py - fabric utility functions for updian
#
#  Copyright (c) 2007-2013 RobHost GmbH <support@robhost.de>
#
#  Author: Dirk Dankhoff <dd@robhost.de>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA
'''Utility functions for handling updian's data structures'''

import fabric.api

class UnknownBackendError(Exception):
    '''Raised when an unsupported backend is used.'''
    def __init__(self, backend):
        super().__init__('Unknown backend: %s' % backend)

def _detect_backend():
    '''Detect package management backend.'''
    supported_backends = {
        '/usr/bin/apt-get': None,
        '/usr/bin/yum': None,
    }

    for backend in supported_backends:
        with fabric.api.settings(ok_ret_codes=[0, 1]):
            p = fabric.api.run('test -x %s' % backend, quiet=True)
        supported_backends[backend] = p.return_code

    available_backends = [x for x in supported_backends if
                          supported_backends[x] == 0]

    # shouldn't really happen, but better be safe then sorry
    # as there are things like apt-rpm around
    if len(available_backends) > 1:
        raise RuntimeError('Auto-detection of package manager returned '
                           'ambiguous results. More than one package '
                           'manager found on %s.' % fabric.api.env.host)

    backend = available_backends[0]

    if backend == '/usr/bin/apt-get':
        return 'apt'
    elif backend == '/usr/bin/yum':
        return 'yum'

    return None

def update_check(backend=None, use_sudo=False):
    '''Check for available package updates using the specified backend.

    When use_sudo is True (default=False) sudo is used to start the
    specified backend.
    If backend is None (the default) auto-detection is performed.

    '''
    driver = fabric.api.sudo if use_sudo else fabric.api.run

    if backend is None:
        backend = _detect_backend()

    if backend == 'apt':
        driver('DEBIAN_FRONTEND=noninteractive apt-get update -qq', quiet=True)
        with fabric.api.settings(ok_ret_codes=[0, 1]):
            ret = driver('DEBIAN_FRONTEND=noninteractive '
                         'apt-get upgrade -s | '
                         'grep Inst')
    elif backend == 'yum':
        with fabric.api.settings(ok_ret_codes=[0, 100]):
            ret = driver('yum check-update -q')
    else:
        raise UnknownBackendError(backend)

    return ret

def upgrade_packages(backend=None,
                     use_sudo=False,
                     allow_unauthenticated_packages = False):
    '''Upgrade all packages using the specified backend.

    When use_sudo is True (default=False) sudo is used to start the
    specified backend.
    If backend is None (the default) auto-detection is performed.

    '''
    driver = fabric.api.sudo if use_sudo else fabric.api.run

    if backend is None:
        backend = _detect_backend()

    if backend == 'apt':
        command = ('PAGER=cat DEBIAN_FRONTEND=noninteractive '
                   'apt-get --yes '
                   '-o Dpkg::Options::="--force-confdef" '
                   '-o Dpkg::Options::="--force-confold" upgrade')
        if allow_unauthenticated_packages:
            command += ' --allow-unauthenticated '
    elif backend == 'yum':
        command = 'yum -y update'
    else:
        raise UnknownBackendError(backend)

    return driver(command, pty=False)

def checkrestart(use_sudo=False):
    '''Issue the checkrestart-command.'''
    driver = fabric.api.sudo if use_sudo else fabric.api.run
    command = 'checkrestart'

    return driver(command)

def make_host_list(serverlist):
    '''Create fabric compatible host list from updian's server list.'''
    host_list = []

    for server in serverlist:
        user = server.user or server.defaults['user']
        port = server.port or server.defaults['port']
        hostspec = '%s@%s:%d' % (user, server.hostname, port)

        host_list.append(hostspec)

    return host_list



# views.py - frontend views
#
#  Copyright (c) 2007-2013 RobHost GmbH <support@robhost.de>
#
#  Author: Dirk Dankhoff <dd@robhost.de>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA
import os
import time

import flask

from flask import flash, redirect, render_template, request, url_for

from . import forms
from . import validation
from .utils import *
from ..frontend import app
from .. import config
from ..serverlist import Server, ServerList

@app.route('/')
def home():
    return list_updates()

def list_updates():
    template_data = dict()

    statfile = os.path.join(config.data_dir, 'statfile')

    template_data['lastupdate'] = formatted_mtime(statfile)

    host_list = list()
    for hostname in sorted(hosts_with_updates()):
        with open(get_data_filename(hostname), 'r') as f:
            count = sum(1 for line in f)
        host = dict(name=hostname,
                    update_count=count,
                    queued=is_queued(hostname))
        host_list.append(host)

    if host_list:
        template_data['updatelist'] = host_list

    server = request.args.get('server', None)
    if server:
        update_details = get_update_details(server)
        if update_details:
            template_data['update_details'] = update_details

    return render_template('index.html', **template_data)

@app.route('/queue/')
@app.route('/queue/list/')
def show_queue():
    template_data = dict()

    statfile = os.path.join(config.data_dir, 'statfile_upd')

    template_data['lastrun'] = formatted_mtime(statfile)

    host_list = list()
    for hostname in sorted(queued_hosts()):
        mtime = formatted_mtime(get_todo_filename(hostname))
        host = dict(name=hostname,
                    queue_date=mtime)
        host_list.append(host)

    if host_list:
        template_data['updatequeue'] = host_list

    return render_template('queue_show.html', **template_data)

@app.route('/queue/all')
def queue_all_hosts():
    for host in hosts_with_updates():
        queue_host(host)

    return redirect(url_for('home'))

@app.route('/queue/add/<hostname>')
def queue_host(hostname):
    todo_file = get_todo_filename(hostname)
    if not os.path.exists(todo_file):
        open(todo_file, 'w').close()

    return redirect(url_for('home'))

@app.route('/queue/remove/<hostname>')
def dequeue_host(hostname):
    # TODO restrict to POST to get CSRF protection
    todo_file = get_todo_filename(hostname)
    if os.path.exists(todo_file):
        os.remove(todo_file)

    return redirect(url_for('show_queue'))

@app.route('/servers/')
@app.route('/servers/list/')
def list_servers():
    serverlist_file = config.serverlist_file
    if not serverlist_file.endswith('.json'):
        flash('Please convert your server.txt to server.json.')

    try:
        serverlist = get_serverlist(serverlist_file)
        serverlist.sort()
    except ValueError as e:
        e = 'There was an error decoding your server list file: %s\n' % e
        e += 'Maybe you forgot to convert it from the old server.txt format?'
        return critical_error(e)
    except IOError:
        flash('There was an error reading the server list file. '
              'Creating empty server list. '
              'It will be saved once you create the first server.')
        serverlist = ServerList()

    template_data = dict(
        serverlist_file=serverlist_file,
        serverlist=serverlist)

    return render_template('server_list.html', **template_data)

@app.route('/servers/edit/',
           defaults={'hostname': None},
           methods=['GET', 'POST'])
@app.route('/servers/edit/<hostname>', methods=['GET', 'POST'])
def edit_server(hostname):
    try:
        serverlist = get_serverlist()
    except IOError:
        serverlist = ServerList()

    if hostname is None:
        server = Server(None)
    else:
        try:
            server_id = serverlist.index(Server(hostname))
            server = serverlist[server_id]
        except ValueError:
            flash('Host not found')
            return redirect(url_for('list_servers'))

    if request.method == 'POST':
        server = forms.server_from_form(request.form)
        errors = validation.validate_server(server)

        if not request.form['host_id'] and server in serverlist:
            # new hostname is already in serverlist
            errors.insert(0, 'Duplicate hostname. '
                          'A host with the name "%s" already exists.' %
                          server.hostname)

        if errors:
            map(flash, errors)
        else:
            if not request.form['host_id']:
                # new entry
                if server in serverlist:
                    flash()
                else:
                    serverlist.append(server)
            else:
                # edit existing entry
                existing_id = serverlist.index(Server(request.form['host_id']))
                serverlist[existing_id] = server

            serverlist.sort()
            serverlist.dump(config.serverlist_file)

            return redirect(url_for('list_servers'))

    return render_template('server_edit.html', host_id=hostname, server=server)

@app.route('/servers/delete', methods=['POST'])
def delete_server():
    serverlist = get_serverlist()

    host_to_remove = request.form['hostname']
    serverlist.remove(Server(host_to_remove))

    serverlist.dump(config.serverlist_file)

    return redirect(url_for('list_servers'))

@app.route('/logs/')
def list_logs():
    template_data = dict()

    statfile = os.path.join(config.data_dir, 'statfile_upd')

    template_data['lastrun'] = formatted_mtime(statfile)

    log_list = list()
    for hostname in sorted(available_logs()):
        mtime = formatted_mtime(get_log_filename(hostname))
        host = dict(name=hostname,
                    update_date=mtime)
        log_list.append(host)

    if log_list:
        template_data['logs'] = log_list

    return render_template('logs_list.html', **template_data)

@app.route('/logs/show/<hostname>')
def show_logfile(hostname):
    if '..' in hostname or hostname.startswith('/'):
        flask.abort(404)

    logfile = get_log_filename(hostname)

    if not os.path.exists(logfile):
        flask.abort(404)

    return flask.send_file(logfile, 'text/plain')

@app.route('/logs/delete/<hostname>')
def delete_logfile(hostname):
    # TODO use POST to get CSRF protection
    logfile = get_log_filename(hostname)
    if not os.path.exists(logfile):
        flask.abort(404)

    os.remove(logfile)

    return redirect(url_for('list_logs'))

@app.route('/logs/delete_all')
def delete_all_logfiles():
    # FIXME: This has the possibility of deleting more than the user has seen
    for hostname in available_logs():
        delete_logfile(hostname)

    return redirect(url_for('list_logs'))

@app.errorhandler(500)
def critical_error(e):
    return render_template('critical_error.html', error=e), 500



# serverlist.py - serverlist handling
#
#  Copyright (c) 2007-2013 RobHost GmbH <support@robhost.de>
#
#  Author: Dirk Dankhoff <dd@robhost.de>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA
'''Updian server list handling'''

from __future__ import print_function

import functools
import json
import operator
import os.path
import sys
import types

FMT_CSV = 'format_csv'
FMT_JSON = 'format_json'

available_formats = [FMT_CSV, FMT_JSON]

class ServerList(list):
    '''Class providing a list interface for updian's server list.

    It also provides convenience methods for loading and dumping the
    list from/to the filesystem.

    '''
    @classmethod
    def from_file(cls, server_file, fmt=available_formats[-1]):
        return read_from_file(server_file, fmt)

    def dump(self, server_file, indent=None):
        def dump_it(slist, fp, indent):
            json.dump(slist, fp, cls=ServerListJSONEncoder, indent=indent)

        if isinstance(server_file, file):
            return dump_it(self, server_file, indent)

        with open(server_file, 'w') as fp:
            return dump_it(self, fp, indent)

    def sort(self):
        super(ServerList, self).sort(key=operator.attrgetter('hostname'))

class ServerListJSONEncoder(json.JSONEncoder):
    '''JSON encoder for Server objects.'''
    def default(self, obj):
        if isinstance(obj, Server):
            return dict(obj)
        return json.JSONEncoder.default(self, obj)

class Server(object):
    defaults = {
        'user': 'root',
        'port': 22,
        'backend': 'apt',
        'gateway': None,
    }

    def __init__(self, hostname, port=None, backend=None, user=None,
                 gateway=None):
        if port and type(port) != types.IntType:
            port = int(port)

        self.hostname = hostname
        self.port = port
        self.backend = backend
        self.user = user
        self.gateway = gateway

    def __iter__(self):
        for attr in self.__dict__:
            if not attr.startswith('_') and getattr(self, attr):
                yield (attr, getattr(self, attr))

    def __repr__(self):
        return repr(dict(self))

    def __eq__(self, other):
        return self.hostname == other.hostname

    def __ne__(self, other):
        return self.hostname != other.hostname

def _file_format_dispatch(function):
    def get_target_function(fmt):
        if fmt not in available_formats:
            raise ValueError("Config file format '%s' not supported." % fmt)

        target_name = '_%s_%s' % (function.__name__, fmt)

        try:
            target_function = globals()[target_name]
        except KeyError:
            raise NotImplementedError("Oops, we forgot to implement '%s' for "
                "format '%s'. Please file a bug report." %
                (function.__name__, fmt))

        if type(target_function) != types.FunctionType:
            raise TypeError("Found '%s' but it's not a function." % target_name)

        return target_function

    if function.__name__.startswith('read_from_'):
        @functools.wraps(function)
        def wrapper(server_file, fmt=available_formats[-1], *args, **kwargs):
            target_function = get_target_function(fmt)

            if isinstance(server_file, file):
                return target_function(server_file, *args, **kwargs)

            with open(server_file, 'r') as fp:
                return target_function(fp, *args, **kwargs)

        return wrapper
    else:
        raise ValueError('file_format_dispatch() only works for '
            'read_from-functions.')

@_file_format_dispatch
def read_from_file():
    '''Read server list from file.'''
    pass

def _read_from_file_format_csv(fp):
    slist = ServerList()

    for line in fp.readlines():
        fields = line.strip().split(':')
        slist.append(Server(*fields))

    return slist

def _read_from_file_format_json(fp):
    def list_entry_hook(dct):
        return Server(**dct)

    deserialized_data = json.load(fp, object_hook=list_entry_hook)

    if not isinstance(deserialized_data, list):
        raise ValueError('Deserialized object was no list.')

    return ServerList(deserialized_data)

def convert_from_csv(old_server_file='server.txt', new_server_file='server.json'):
    if not os.path.exists(old_server_file):
        print('ERROR: %s not found' % old_server_file, file=sys.stderr)

    slist = read_from_file(old_server_file, FMT_CSV)

    if os.path.exists(new_server_file):
        prompt = ('Output file "%s" already exists. Overwrite? [Y/n] '
                  % new_server_file)
        answer = raw_input(prompt).lower()

        if answer in ['n', 'no']:
            new_server_file += '.new'

    slist.dump(new_server_file, 2)

    print('Converted %s to %s.' % (old_server_file, new_server_file))



# fabfile.py - fabric deployment script
#
#  Copyright (c) 2007-2013 RobHost GmbH <support@robhost.de>
#
#  Author: Dirk Dankhoff <dd@robhost.de>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA
'''Fabfile containing all operations needed to deploy Updian locally.'''
from __future__ import print_function

import os
import shutil
import sys

import fabric.colors
from fabric.api import task, local, cd
from fabric.contrib.console import confirm

venv_path = os.path.join(os.getcwd(), 'updian-venv')

def warning(text, prefix='Warning:'):
    print(fabric.colors.yellow(prefix + ' ', bold=True) + text)

def make_virtual_environment():
    local('virtualenv --no-site-packages %s' % venv_path)

def activate_virtual_environment():
    activate_this = os.path.join(venv_path, 'bin/activate_this.py')
    execfile(activate_this, dict(__file__=activate_this))

def install_updian():
    try:
        import updian
    except ImportError:
        local('pip install updian')

def make_runtime_directories():
    mode = 0750
    os.mkdir('log/', mode)
    os.mkdir('data/', mode)
    os.mkdir('todo/', mode)

def copy_frontend_data():
    import updian
    base_dir = os.path.join(os.path.dirname(updian.__file__), 'frontend')
    cur_dir = os.getcwd()

    for d in ['static', 'templates']:
        src_dir = os.path.join(base_dir, d)
        dst_dir = os.path.join(cur_dir, d)
        shutil.copytree(src_dir, dst_dir)

def download_frontend_dependencies():
    dep_urls = [
        'http://code.jquery.com/jquery-2.0.2.min.js',
        'http://ajax.aspnetcdn.com/ajax/jquery.dataTables/1.9.4/jquery.dataTables.min.js',
    ]

    for dep_url in dep_urls:
        local('wget "%s" -P static/js' % dep_url)

def initialize_config():
    print('Creating user for basic authentication...')
    local('updiancmd setpw')
    print('Initializing configuration...')
    local('updiancmd init_cfg')

    if not os.path.exists('updian.wsgi'):
        with open('updian.wsgi', 'w') as wsgi_file:
            print('from updian.frontend import app as application',
                  file=wsgi_file)

def completion_message():
    print()
    print('Updian is now installed.')
    print('You should now edit the file `config.ini` to your liking.')
    print('Afterwards you can use `updiancmd runserver [IP]` to start the '
          'builtin Flask web server or configure your webserver to serve the '
          'wsgi application from the updian.wsgi in this directory.')
    print()
    warning('Only use the builtin web server if you serve Updian using a '
            'secure local network interface.', prefix='Note:')

@task
def make_dist():
    base_dir = os.path.dirname(os.path.dirname(__file__))
    with cd(os.path.abspath(base_dir)):
        local('python setup.py sdist')

@task(default=True)
def deploy():
    if 'VIRTUAL_ENV' not in os.environ:
        make_virtual_environment()
        activate_virtual_environment()
    else:
        warning('You are already inside a virtual environment (%s). '
                'Updian will be installed inside it.' %
                os.environ['VIRTUAL_ENV'])
        install_to_current_venv = confirm('Is this correct?')
        if not install_to_current_venv:
            create_new_venv = confirm('Should we create a new virtual '
                                      'environment?')
            if create_new_venv:
                make_virtual_environment()
                activate_virtual_environment()
            else:
                print('Cannot continue. Aborting.')
                sys.exit(1)

    install_updian()
    make_runtime_directories()
    copy_frontend_data()
    download_frontend_dependencies()
    initialize_config()
    completion_message()



# forms.py - functions for frontend form handling
#
#  Copyright (c) 2007-2013 RobHost GmbH <support@robhost.de>
#
#  Author: Dirk Dankhoff <dd@robhost.de>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA
from ..serverlist import Server

def server_from_form(form):
    return Server(hostname=form['hostname'],
                  port=form['port'] or None,
                  backend=form['backend'] or None,
                  user=form['user'] or None,
                  gateway=form['gateway'] or None)



# convert_serverlist.py - command to convert old server lists
#
#  Copyright (c) 2007-2013 RobHost GmbH <support@robhost.de>
#
#  Author: Dirk Dankhoff <dd@robhost.de>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA
from __future__ import print_function

import os

from updian import config, serverlist

sl = config.serverlist_file
if sl.endswith('.txt'):
    print('Converting...')
    slnew = sl.replace('.txt', '.json')
    if os.path.exists(slnew):
        slnew += '.convnew'
    serverlist.convert_from_csv(sl, slnew)
    print('Saving to file %s.' % slnew)
    print('Now please check the converted file and adjust your config file accordingly')
else:
    print('Server list was apparently already converted to JSON. Aborting.')



# passwd.py - passwd handling functions
#
#  Copyright (c) 2007-2013 RobHost GmbH <support@robhost.de>
#
#  Author: Dirk Dankhoff <dd@robhost.de>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA
from __future__ import print_function

import crypt
import getpass
import os
import random
import string
import sys

SALT_CHARACTERS = string.ascii_letters + string.digits + '/.'
SALT_PREFIX = '$6$'

passwd_file = 'updian.passwd'

def gensalt(length):
    if length < 8 or length > 16:
        raise ValueError('Salt length must be between 8 and 16.')

    salt = ''.join([random.choice(SALT_CHARACTERS) for _ in range(length)])

    return salt

def hashpw(password, salt):
    if not salt.startswith(SALT_PREFIX):
        for c in salt:
            if c not in SALT_CHARACTERS:
                raise ValueError('Salt contains invalid character: %s' % c)
        salt = SALT_PREFIX + salt + '$'

    return crypt.crypt(password, salt)

def line_contains_username(line, username):
    return line.startswith(username + ':$')

def get_user_from_passwd(username):
    with open(passwd_file, 'r') as f:
        for line in f:
            if line_contains_username(line, username):
                return line.strip().rsplit(':', 1)
    return None

def set_user_password(username, password):
    if ':$' in username:
        raise ValueError('Username must not contain ":$".')

    if not os.path.exists(passwd_file):
        open(passwd_file, 'w').close()

    user = get_user_from_passwd(username)
    pwhash = hashpw(password, gensalt(16))

    new_user_line = username + ':' + pwhash + '\n'

    if user is None:
        with open(passwd_file, 'a') as f:
            f.write(new_user_line)
    else:
        with open(passwd_file, 'r') as old_passwd:
            lines = old_passwd.readlines()
        with open(passwd_file, 'w') as f:
            for l in lines:
                if line_contains_username(l, username):
                    l = new_user_line
                f.write(l)

def set_interactively():
    if not os.path.exists(passwd_file):
        open(passwd_file, 'w').close()

    username = raw_input('Username: ')

    user = get_user_from_passwd(username)
    if user is not None:
        current_pw = raw_input('Current password: ')
        if hashpw(current_pw, user[1]) != user[1]:
            print('Current password does not match.')
            sys.exit(1)

    password = getpass.getpass('New password: ')
    password2 = getpass.getpass('Repeat new password: ')

    if password != password2:
        print('Passwords do not match')
        sys.exit(1)

    set_user_password(username, password)

if __name__ == '__main__':
    set_interactively()



# collect.py - cronjob functionality for collecting pending updates
#
#  Copyright (c) 2007-2013 RobHost GmbH <support@robhost.de>
#
#  Author: Dirk Dankhoff <dd@robhost.de>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA
'''Check hosts for updates and send reports.

Module variables:
    update_mail_subject -- Subject of the summary email. The placeholder
                           `{count}` is replaced with the total update
                           count.
    update_mail_subject -- Body of the summary email. The following
                           placeholders are recognized and replaced:
                             `{count}`:
                                 Total update count.

                             `{update_list}`:
                                 List of update counts per server
                                 separated by line breaks.

                             `{updian_uri}`:
                                 URI of Updian's web-interface as
                                 globally configured.

'''

from __future__ import print_function

import email.mime.text
import glob
import os
import smtplib
import time

from fabric.api import parallel, execute, env, settings, hide

from .. import config
from ..fabric_utils import make_host_list, update_check

__all__ = ['update_mail_subject', 'update_mail_text', 'collect_update_data']

update_mail_subject = '[updian] {count} servers with updates pending!'
update_mail_text = '''
Updian has detected that the following servers have pending updates:

{update_list}

You can manage these updates at {updian_uri}

Regards,
updian on base
'''

@parallel(pool_size=config.concurrency)
def check_for_update(metadata_mapping):
    '''Fabric task that checks hosts for pending updates.'''
    metadata = metadata_mapping[env.host]
    backend = metadata.backend
    use_sudo = (env.user != 'root')

    env.gateway = metadata.gateway or metadata.defaults['gateway']

    if backend is None and not config.autodetect_backend:
        backend = metadata.defaults['backend']
        print('No backend given for %s and auto-detection is disabled. '
              'Defaulting to %s' % (env.host, backend))

    print('Query: %s, Port: %s, Engine: %s, Gateway: %s' %
          (env.host, env.port, backend, env.gateway))

    env.shell = '/bin/bash -c'

    s = update_check(backend, use_sudo)
    result = s.stdout.replace('\r', '')

    update_count = len(result.splitlines())

    if update_count:
        filename = os.path.join(config.data_dir, '%s.txt' % env.host)
        with open(filename, 'w') as outfile:
            outfile.write(result)

    return update_count

def clear_datadir():
    '''Delete all .txt-files from the data directory.'''
    for f in glob.glob(os.path.join(config.data_dir, '*.txt')):
        os.remove(f)

def collect_update_data(serverlist):
    '''Execute update checks on all hosts in the global server list.

    An email summarizing the pending updates is sent if mailing is
    enabled in the global configuration.

    '''
    clear_datadir()

    metadata_mapping = dict(((s.hostname, s) for s in serverlist))

    with hide('everything'):
        updates = execute(check_for_update,
                          metadata_mapping,
                          hosts=make_host_list(serverlist))

    statfile = os.path.join(config.data_dir, 'statfile')
    open(statfile, 'w').close()

    if config.mail_active:
        update_list = ['%s: %d pending updates' %
                       (host, updates[host]) for host in updates]

        mailtext = update_mail_text.format(
            count=len(updates),
            update_list='\n'.join(update_list),
            updian_uri=config.updian_uri)

        mail = email.mime.text.MIMEText(mailtext)
        mail['Subject'] = update_mail_subject.format(count=len(updates))
        mail['From'] = 'updian <%s>' % (config.mail_from)
        mail['To'] = config.mail_to

        smtp_server = smtplib.SMTP('localhost')
        smtp_server.sendmail(config.mail_from, [config.mail_to],
                             mail.as_string())
        smtp_server.quit()

        print('Mail sent to', config.mail_to)

if __name__ == '__main__':
    from ..serverlist import ServerList

    serverlist = ServerList.from_file(config.serverlist_file)

    collect_update_data(serverlist)



# validation.py - validation functions for entity models
#
#  Copyright (c) 2007-2013 RobHost GmbH <support@robhost.de>
#
#  Author: Dirk Dankhoff <dd@robhost.de>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA
import re

def validate_server(server):
    errors = []

    if not server.hostname:
        errors.append('Hostname must not be empty.')

    if server.port is not None and (server.port < 1 or server.port > 65535):
        errors.append('Port has to be an integer between 1 and 65535 if set.')

    if server.user and re.search('[@\s]', server.user):
        errors.append('Username must not include @-signs or whitespace.')

    if (server.gateway and
        not re.match('^([^@:\s]+@)?([^@:\s]+)(:\d+)?$', server.gateway)):
        errors.append('Gateway must be given in format "user@host:port". '
                      'User and port are optional.')

    return errors



# basicauth.py - Flask app wrapper to provide global http basic auth
#
#  Copyright (c) 2007-2013 RobHost GmbH <support@robhost.de>
#
#  Author: Dirk Dankhoff <dd@robhost.de>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA
import os

import flask.ext.basicauth

from .. import passwd

class BasicAuth(flask.ext.basicauth.BasicAuth):
    def __init__(self, app):
        if os.path.exists(passwd.passwd_file):
            app.config['BASIC_AUTH_FORCE'] = True
        super(BasicAuth, self).__init__(app)

    def check_credentials(self, username, password):
        user = passwd.get_user_from_passwd(username)

        if user is not None:
            return passwd.hashpw(password, user[1]) == user[1]

        return False



import flask
from flaskext.csrf import csrf
from .basicauth import BasicAuth
from .. import config

app = flask.Flask(__name__)
app.secret_key = config.secret_key

csrf(app)

basic_auth = BasicAuth(app)

from . import views



# utils.py - utility functions
#
#  Copyright (c) 2007-2013 RobHost GmbH <support@robhost.de>
#
#  Author: Dirk Dankhoff <dd@robhost.de>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA
import os.path
import time

from glob import glob

from .. import config
from ..serverlist import ServerList

def get_data_filename(hostname):
    return os.path.join(config.data_dir, hostname) + '.txt'

def has_updates(hostname):
    return os.path.exists(get_data_filename(hostname))

def get_todo_filename(hostname):
    return os.path.join(config.todo_dir, hostname) + '.txt'

def is_queued(hostname):
    return os.path.exists(get_todo_filename(hostname))

def get_log_filename(hostname):
    return os.path.join(config.log_dir, hostname) + '.log'

def hostlist_decorator(basedir, extension='txt'):
    def decorator(fun):
        def get_hostlist():
            for host_file in glob(os.path.join(basedir, '*.' + extension)):
                host = os.path.splitext(os.path.basename(host_file))[0]
                yield host

        return get_hostlist
    return decorator

@hostlist_decorator(config.data_dir)
def hosts_with_updates():
    pass

@hostlist_decorator(config.todo_dir)
def queued_hosts():
    pass

@hostlist_decorator(config.log_dir, 'log')
def available_logs():
    pass

def get_update_details(hostname):
    if not has_updates(hostname):
        return None

    with open(get_data_filename(hostname), 'r') as f:
        update_list = [line.replace('Inst ', '') for line in f]

    return dict(host=hostname,
                queued=is_queued(hostname),
                updates=update_list)

def get_serverlist(serverlist_file = None):
    if serverlist_file is None:
        serverlist_file = config.serverlist_file

    return ServerList.from_file(serverlist_file)

def formatted_mtime(filename):
    timeformat = '%B %d %Y %H:%M'
    try:
        mtime = os.path.getmtime(filename)
    except os.error:
        if os.path.exists(filename):
            raise
        else:
            return None

    return time.strftime(timeformat, time.localtime(mtime))



# update.py - cronjob functionality for installing pending updates
#
#  Copyright (c) 2007-2013 RobHost GmbH <support@robhost.de>
#
#  Author: Dirk Dankhoff <dd@robhost.de>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA
'''Functions to run updates via updian.'''

from __future__ import print_function

import glob
import os
import random
import time

from fabric.api import parallel, execute, env, settings, hide

from .. import config
from ..fabric_utils import *

__all__ = ['execute_queued_updates']

@parallel(pool_size=config.concurrency)
def do_update(metadata_mapping):
    '''Fabric task that updates packages and logs the output.'''
    metadata = metadata_mapping[env.host]
    backend = metadata.backend
    use_sudo = (env.user != 'root')

    env.gateway = metadata.gateway or metadata.defaults['gateway']

    if backend is None and not config.autodetect_backend:
        backend = metadata.defaults['backend']
        print('No backend given for %s and auto-detection is disabled. '
              'Defaulting to %s' % (env.host, backend))

    print('Host: %s, Port: %s, Engine: %s, Gateway: %s' %
          (env.host, env.port, backend, env.gateway))

    env.shell = '/bin/bash -c'

    s = upgrade_packages(backend, use_sudo,
                         config.allow_unauthenticated_packages)

    log_entry = '{host} ({date}):\n\n{output}\n\n####################\n\n'
    log_entry = log_entry.format(host=env.host,
                                 date=time.strftime('%b %d, %H:%M'),
                                 output=s.stdout.replace('\r', ''))

    logfile = os.path.join(config.log_dir, '%s.log' % env.host)
    with open(logfile, 'a+') as log:
        log.write(log_entry)
    os.chmod(logfile, 0666)

    if backend == 'apt':
        chkrst_res = checkrestart(use_sudo)
        output = chkrst_res.replace('\r', '').strip()

        if output != 'Found 0 processes using old versions of upgraded files':
            chkrst_logfile = '%s_checkrestart.log' % (env.host)
            chkrst_logfile = os.path.join(config.log_dir, chkrst_logfile)
            with open(chkrst_logfile, 'a+') as chkrst_log:
                chkrst_log.write(output)
            os.chmod(chkrst_logfile, 0666)

    queuefile = os.path.join(config.todo_dir, '%s.txt' % env.host)
    os.remove(queuefile)

    datafile = os.path.join(config.data_dir, '%s.txt' % env.host)
    os.remove(datafile)

def execute_queued_updates(serverlist):
    '''Execute updates on all queued hosts (in the todo directory).'''
    todo_files = glob.glob(os.path.join(config.todo_dir, '*.txt'))
    host_queue = [os.path.splitext(os.path.basename(f))[0] for f in todo_files]

    if not host_queue:
        print('No updates in queue (%s) ...' % (time.strftime('%H:%M')))
        return

    serverlist = filter(lambda s: s.hostname in host_queue, serverlist)
    random.shuffle(serverlist)
    metadata_mapping = dict(((s.hostname, s) for s in serverlist))

    with hide('everything'):
        return_vals = execute(do_update,
                              metadata_mapping,
                              hosts=make_host_list(serverlist))

    statfile = os.path.join(config.data_dir, 'statfile_upd')
    open(statfile, 'w').close()

    print('Made %d updates (%s) ...' % (len(serverlist),
                                        time.strftime('%H:%M')))

if __name__ == '__main__':
    from ..serverlist import ServerList

    serverlist = ServerList.from_file(config.serverlist_file)

    execute_queued_updates(serverlist)



# config.py - global configuration
#
#  Copyright (c) 2007-2013 RobHost GmbH <support@robhost.de>
#
#  Author: Dirk Dankhoff <dd@robhost.de>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA
'''Updian configuration initialization and config file parsing'''

from ConfigParser import RawConfigParser

# configuration defaults
option_defaults = dict(
    serverlist_file = 'server.json',
    data_dir = 'data',
    todo_dir = 'todo',
    log_dir = 'log',
    updian_uri = 'http://192.168.0.254/updian',
    mail_active = True,
    mail_to = 'root@localhost',
    mail_from = 'updian@localhost',
    concurrency = 20,
    autodetect_backend = True,
    allow_unauthenticated_packages = False,
    secret_key = 'SECRET')

main_section = 'UPDIAN'

example_config = (
'''[%s]
# file containing the server list
serverlist_file = server.json

# secret key used for CSRF and cookie encryption
secret_key = {secret_key}

# url to your installation (used for hyperlinking in mails)
updian_url = http://192.168.0.254/updian/

# send infomails: true|false
mail_active = true
# recipient of infomails, should be your valid email
mail_to     = server@domain.tld
# sender of informails, defaults to updian@localhost, have a look at "hostname -f"
mail_from   = updian@domain.tld

# concurrency options (how man processes to for for update command)
concurrency = 20

# backend auto-detection
# if this is set to false, the default backend (apt) will be used for hosts w/o an explicit backend configuration
autodetect_backend = true

# allow unauthenticated packages
allow_unauthenticated_packages = false
''' % (main_section))

def write_example_config(filename='config.ini'):
    '''Write a configuration example to a file.'''
    from random import choice
    from string import ascii_letters, digits

    secret_len = 32
    secret = ''.join([choice(ascii_letters + digits) for _ in
                      range(secret_len)])

    with open(filename, 'w') as fp:
        fp.write(example_config.format(secret_key = secret))

def read_config_ini(filename='config.ini'):
    '''Read config file and update global configuration variables.'''
    cp = RawConfigParser()

    found = cp.read(filename)

    if len(found) == 0:
        # TODO no config files loaded, using defaults only
        pass

    def get_option(option):
        default = option_defaults[option]

        if not cp.has_option(main_section, option):
            return default

        if isinstance(default, bool):
            return cp.getboolean(main_section, option)
        elif isinstance(default, int):
            return cp.getint(main_section, option)
        elif isinstance(default, float):
            return cp.getfloat(main_section, option)

        return cp.get(main_section, option)

    module_vars = globals()

    for option in option_defaults:
        v = get_option(option)
        module_vars[option] = v

# initialize on import
read_config_ini()



# setup.py - python distribute install script
#
#  Copyright (c) 2007-2013 RobHost GmbH <support@robhost.de>
#
#  Author: Dirk Dankhoff <dd@robhost.de>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA
import os
import sys
from setuptools import setup, find_packages

if sys.version_info[:2] < (2, 6) or sys.version_info[0] > 2:
    msg = ("UPDIAN requires Python 2.6 or later but does not work on "
           "any version of Python 3. You are using version %s. Please "
           "install using a supported version." % sys.version)
    sys.stderr.write(msg)
    sys.exit(1)

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

requires = read('requirements.txt').splitlines()

CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'Intended Audience :: System Administrators',
    'Natural Language :: English',
    'Operating System :: POSIX :: Linux',
    'Topic :: System :: Monitoring',
    'Topic :: System :: Systems Administration',
    ]

setup(
    name = 'UPDIAN',
    version = '0.6',
    license = 'GPLv2+ <http://www.gnu.org/licenses/gpl-2.0.en.html>',
    url = 'http://www.robhost.de/updian/',
    description = "A program for monitoring and executing updates on remote systems",
    long_description = read('README.rst'),
    classifiers = CLASSIFIERS,
    author = "Robert Klikics",
    author_email = "rk@robhost.de",
    maintainer = "Dirk Dankhoff",
    maintainer_email = "dd@robhost.de",
    install_requires = requires,
    packages = find_packages(),
    scripts = ['updiancmd'],
    include_package_data = True,
    zip_safe = False,
    )
